"""
Unified Configuration System for Outerbounds Apps

This module provides a type-safe, declarative configuration system that serves as the
single source of truth for app configuration. It automatically generates CLI options,
handles config file parsing, and manages field merging behavior.

No external dependencies required - uses only Python standard library.
"""


import os
import json
from typing import Any, Dict, List, Optional, Union, Type
import re

from .config_utils import (
    ConfigField,
    ConfigMeta,
    JsonFriendlyKeyValuePairType,
    PureStringKVPairType,
    CommaSeparatedListType,
    FieldBehavior,
    CLIOption,
    config_meta_to_dict,
    merge_field_values,
    apply_defaults,
    populate_config_recursive,
    validate_config_meta,
    validate_required_fields,
    ConfigValidationFailedException,
    commit_owner_names_across_tree,
)


class AuthType:
    BROWSER = "Browser"
    API = "API"

    @classmethod
    def choices(cls):
        return [cls.BROWSER, cls.API]


class UnitParser:
    UNIT_FREE_REGEX = r"^\d+$"

    metrics = {
        "memory": {
            "default_unit": "Mi",
            "requires_unit": True,  # if a Unit free value is provided then we will add the default unit to it.
            # Regex to match values with units (e.g., "512Mi", "4Gi", "1024Ki")
            "correct_unit_regex": r"^\d+(\.\d+)?(Ki|Mi|Gi|Ti|Pi|Ei)$",
        },
        "cpu": {
            "default_unit": None,
            "requires_unit": False,  # if a Unit free value is provided then we will not add the default unit to it.
            # Accepts values like 400m, 4, 0.4, 1000n, etc.
            # Regex to match values with units (e.g., "400m", "1000n", "2", "0.5")
            "correct_unit_regex": r"^(\d+(\.\d+)?(m|n)?|\d+(\.\d+)?)$",
        },
        "disk": {
            "default_unit": "Mi",
            "requires_unit": True,  # if a Unit free value is provided then we will add the default unit to it.
            # Regex to match values with units (e.g., "100Mi", "1Gi", "500Ki")
            "correct_unit_regex": r"^\d+(\.\d+)?(Ki|Mi|Gi|Ti|Pi|Ei)$",
        },
        "gpu": {
            "default_unit": None,
            "requires_unit": False,
            # Regex to match values with units (usually just integer count, e.g., "1", "2")
            "correct_unit_regex": r"^\d+$",
        },
    }

    def __init__(self, metric_name: str):
        self.metric_name = metric_name

    def validate(self, value: str):
        if re.match(self.metrics[self.metric_name]["correct_unit_regex"], value):
            return True
        return False

    def process(self, value: str):
        value = str(value)
        if self.metrics[self.metric_name]["requires_unit"]:
            if re.match(self.UNIT_FREE_REGEX, value):
                # This means the value is unit free and we need to add the default unit to it.
                value = "%s%s" % (
                    value.strip(),
                    self.metrics[self.metric_name]["default_unit"],
                )
                return value

        return value

    def parse(self, value: Union[str, None]):
        if value is None:
            return None
        return self.process(value)

    @staticmethod
    def validation_wrapper_fn(
        metric_name: str,
    ):
        def validation_fn(value: str):
            if value is None:
                return True
            field_info = ResourceConfig._get_field(ResourceConfig, metric_name)  # type: ignore
            parser = UnitParser(metric_name)
            validation = parser.validate(value)
            if not validation:
                raise ConfigValidationFailedException(
                    field_name=metric_name,
                    field_info=field_info,
                    current_value=value,
                    message=f"Invalid value for `{metric_name}`. Must be of the format {parser.metrics[metric_name]['correct_unit_regex']}.",
                )
            return validation

        return validation_fn


class BasicValidations:
    def __init__(self, config_meta_class, field_name):
        self.config_meta_class = config_meta_class
        self.field_name = field_name

    def _get_field(self):
        return self.config_meta_class._get_field(self.config_meta_class, self.field_name)  # type: ignore

    def enum_validation(self, enums: List[str], current_value):
        if current_value not in enums:
            raise ConfigValidationFailedException(
                field_name=self.field_name,
                field_info=self._get_field(),
                current_value=current_value,
                message=f"Configuration field {self.field_name} has invalid value {current_value}. Value must be one of: {'/'.join(enums)}",
            )
        return True

    def range_validation(self, min_value, max_value, current_value):
        if current_value < min_value or current_value > max_value:
            raise ConfigValidationFailedException(
                field_name=self.field_name,
                field_info=self._get_field(),
                current_value=current_value,
                message=f"Configuration field {self.field_name} has invalid value {current_value}. Value must be between {min_value} and {max_value}",
            )
        return True

    def length_validation(self, max_length, current_value):
        if len(current_value) > max_length:
            raise ConfigValidationFailedException(
                field_name=self.field_name,
                field_info=self._get_field(),
                current_value=current_value,
                message=f"Configuration field {self.field_name} has invalid value {current_value}. Value must be less than {max_length}",
            )
        return True

    def regex_validation(self, regex, current_value):
        if not re.match(regex, current_value):
            raise ConfigValidationFailedException(
                field_name=self.field_name,
                field_info=self._get_field(),
                current_value=current_value,
                message=f"Configuration field {self.field_name} has invalid value {current_value}. Value must match regex {regex}",
            )
        return True


class ResourceConfig(metaclass=ConfigMeta):
    """Resource configuration for the app."""

    # TODO: Add Unit Validation/Parsing Support for the Fields.
    cpu = ConfigField(
        default="1",
        cli_meta=CLIOption(
            name="cpu",
            cli_option_str="--cpu",
            help="CPU resource request and limit.",
        ),
        field_type=str,
        example="500m",
        validation_fn=UnitParser.validation_wrapper_fn("cpu"),
        parsing_fn=UnitParser("cpu").parse,
    )
    memory = ConfigField(
        default="4Gi",
        cli_meta=CLIOption(
            name="memory",
            cli_option_str="--memory",
            help="Memory resource request and limit.",
        ),
        field_type=str,
        example="512Mi",
        validation_fn=UnitParser.validation_wrapper_fn("memory"),
        parsing_fn=UnitParser("memory").parse,
    )
    gpu = ConfigField(
        cli_meta=CLIOption(
            name="gpu",
            cli_option_str="--gpu",
            help="GPU resource request and limit.",
        ),
        field_type=str,
        example="1",
        validation_fn=UnitParser.validation_wrapper_fn("gpu"),
        parsing_fn=UnitParser("gpu").parse,
    )
    disk = ConfigField(
        default="20Gi",
        cli_meta=CLIOption(
            name="disk",
            cli_option_str="--disk",
            help="Storage resource request and limit.",
        ),
        field_type=str,
        example="1Gi",
        validation_fn=UnitParser.validation_wrapper_fn("disk"),
        parsing_fn=UnitParser("disk").parse,
    )


class HealthCheckConfig(metaclass=ConfigMeta):
    """Health check configuration."""

    enabled = ConfigField(
        default=False,
        cli_meta=CLIOption(
            name="health_check_enabled",
            cli_option_str="--health-check-enabled",
            help="Whether to enable health checks.",
            is_flag=True,
        ),
        field_type=bool,
        example=True,
    )
    path = ConfigField(
        cli_meta=CLIOption(
            name="health_check_path",
            cli_option_str="--health-check-path",
            help="The path for health checks.",
        ),
        field_type=str,
        example="/health",
    )
    initial_delay_seconds = ConfigField(
        cli_meta=CLIOption(
            name="health_check_initial_delay",
            cli_option_str="--health-check-initial-delay",
            help="Number of seconds to wait before performing the first health check.",
        ),
        field_type=int,
        example=10,
    )
    period_seconds = ConfigField(
        cli_meta=CLIOption(
            name="health_check_period",
            cli_option_str="--health-check-period",
            help="How often to perform the health check.",
        ),
        field_type=int,
        example=30,
    )


class AuthConfig(metaclass=ConfigMeta):
    """Authentication configuration."""

    type = ConfigField(
        default=AuthType.BROWSER,
        cli_meta=CLIOption(
            name="auth_type",
            cli_option_str="--auth-type",
            help="The type of authentication to use for the app.",
            choices=AuthType.choices(),
        ),
        field_type=str,
        example="Browser",
    )
    public = ConfigField(
        default=True,
        cli_meta=CLIOption(
            name="auth_public",
            cli_option_str="--public-access/--private-access",
            help="Whether the app is public or not.",
            is_flag=True,
        ),
        field_type=bool,
        example=True,
    )

    @staticmethod
    def validate(auth_config: "AuthConfig"):
        if auth_config.type is None:
            return True
        return BasicValidations(AuthConfig, "type").enum_validation(
            AuthType.choices(), auth_config.type
        )


class ScalingPolicyConfig(metaclass=ConfigMeta):
    """
    Policies for autoscaling replicas. Available policies:
    - Request based Autoscaling (rpm)
    """

    # TODO Change the defaulting if we have more autoscaling policies.
    rpm = ConfigField(
        field_type=int,
        # TODO: Add a little more to the docstring where we explain the behavior.
        cli_meta=CLIOption(
            name="scaling_rpm",
            cli_option_str="--scaling-rpm",
            help=(
                "Scale up replicas when the requests per minute crosses this threshold. "
                "If nothing is provided and the replicas.max and replicas.min is set then  "
                "the default rpm would be 60."
            ),
        ),
        default=60,
    )


class ReplicaConfig(metaclass=ConfigMeta):
    """Replica configuration."""

    fixed = ConfigField(
        cli_meta=CLIOption(
            name="fixed_replicas",
            cli_option_str="--fixed-replicas",
            help="The fixed number of replicas to deploy the app with. If min and max are set, this will raise an error.",
        ),
        field_type=int,
        example=1,
    )

    min = ConfigField(
        cli_meta=CLIOption(
            name="min_replicas",
            cli_option_str="--min-replicas",
            help="The minimum number of replicas to deploy the app with.",
        ),
        field_type=int,
        example=1,
    )
    max = ConfigField(
        cli_meta=CLIOption(
            name="max_replicas",
            cli_option_str="--max-replicas",
            help="The maximum number of replicas to deploy the app with.",
        ),
        field_type=int,
        example=10,
    )

    scaling_policy = ConfigField(
        cli_meta=None,
        field_type=ScalingPolicyConfig,
        help=(
            "Scaling policy defines the the metric based on which the replicas will horizontally scale. "
            "If min and max replicas are set and are not the same, then a scaling policy will be applied. "
            "Default scaling policies can be 60 rpm (ie 1 rps). "
        ),
    )

    @staticmethod
    def defaults(replica_config: "ReplicaConfig"):
        if all(
            [
                replica_config.min is None,
                replica_config.max is None,
                replica_config.fixed is None,
            ]
        ):
            # if nothing is set then set
            replica_config.fixed = 1
        elif replica_config.min is not None and replica_config.max is None:
            replica_config.max = replica_config.min

        return

    @staticmethod
    def validate(replica_config: "ReplicaConfig"):
        both_min_max_set = (
            replica_config.min is not None and replica_config.max is not None
        )
        fixed_set = replica_config.fixed is not None
        max_is_set = replica_config.max is not None
        min_is_set = replica_config.min is not None
        any_min_max_set = (
            replica_config.min is not None or replica_config.max is not None
        )

        def _greater_than_equals_zero(x):
            return x is not None and x >= 0

        if both_min_max_set and replica_config.min > replica_config.max:  # type: ignore
            raise ConfigValidationFailedException(
                field_name="min",
                field_info=replica_config._get_field("min"),  # type: ignore
                current_value=replica_config.min,
                message="Min replicas cannot be greater than max replicas",
            )
        if fixed_set and any_min_max_set:
            raise ConfigValidationFailedException(
                field_name="fixed",
                field_info=replica_config._get_field("fixed"),  # type: ignore
                current_value=replica_config.fixed,
                message="Fixed replicas cannot be set when min or max replicas are set",
            )

        if max_is_set and not min_is_set:
            raise ConfigValidationFailedException(
                field_name="min",
                field_info=replica_config._get_field("min"),  # type: ignore
                current_value=replica_config.min,
                message="If max replicas is set then min replicas must be set too.",
            )

        if fixed_set and replica_config.fixed < 0:  # type: ignore
            raise ConfigValidationFailedException(
                field_name="fixed",
                field_info=replica_config._get_field("fixed"),  # type: ignore
                current_value=replica_config.fixed,
                message="Fixed replicas cannot be less than 0",
            )

        if min_is_set and not _greater_than_equals_zero(replica_config.min):
            raise ConfigValidationFailedException(
                field_name="min",
                field_info=replica_config._get_field("min"),  # type: ignore
                current_value=replica_config.min,
                message="Min replicas cannot be less than 0",
            )

        if max_is_set and not _greater_than_equals_zero(replica_config.max):
            raise ConfigValidationFailedException(
                field_name="max",
                field_info=replica_config._get_field("max"),  # type: ignore
                current_value=replica_config.max,
                message="Max replicas cannot be less than 0",
            )
        return True


def more_than_n_not_none(n, *args):
    return sum(1 for arg in args if arg is not None) > n


class DependencyConfig(metaclass=ConfigMeta):
    """Dependency configuration."""

    from_requirements_file = ConfigField(
        cli_meta=CLIOption(
            name="dep_from_requirements",
            cli_option_str="--dep-from-requirements",
            help="The path to the requirements.txt file to attach to the app.",
        ),
        field_type=str,
        behavior=FieldBehavior.NOT_ALLOWED,
        example="requirements.txt",
    )
    from_pyproject_toml = ConfigField(
        cli_meta=CLIOption(
            name="dep_from_pyproject",
            cli_option_str="--dep-from-pyproject",
            help="The path to the pyproject.toml file to attach to the app.",
        ),
        field_type=str,
        behavior=FieldBehavior.NOT_ALLOWED,
        example="pyproject.toml",
    )
    python = ConfigField(
        cli_meta=CLIOption(
            name="python",
            cli_option_str="--python",
            help="The Python version to use for the app.",
        ),
        field_type=str,
        behavior=FieldBehavior.UNION,
        example="3.10",
    )
    pypi = ConfigField(
        cli_meta=CLIOption(
            name="pypi",  # TODO: Can set CLI meta to None
            cli_option_str="--pypi",
            help="A dictionary of pypi dependencies to attach to the app. The key is the package name and the value is the version.",
            hidden=True,  # Complex structure, better handled in config file
        ),
        field_type=dict,
        behavior=FieldBehavior.NOT_ALLOWED,
        example={"numpy": "1.23.0", "pandas": ""},
    )
    conda = ConfigField(
        cli_meta=CLIOption(  # TODO: Can set CLI meta to None
            name="conda",
            cli_option_str="--conda",
            help="A dictionary of conda dependencies to attach to the app. The key is the package name and the value is the version.",
            hidden=True,  # Complex structure, better handled in config file
        ),
        field_type=dict,
        behavior=FieldBehavior.NOT_ALLOWED,
        example={"numpy": "1.23.0", "pandas": ""},
    )

    @staticmethod
    def validate(dependency_config: "DependencyConfig"):
        # You can either have from_requirements_file or from_pyproject_toml or python with pypi or conda
        # but not more than one of them.
        if more_than_n_not_none(
            1,
            dependency_config.from_requirements_file,
            dependency_config.from_pyproject_toml,
        ):
            raise ConfigValidationFailedException(
                field_name="from_requirements_file",
                field_info=dependency_config._get_field("from_requirements_file"),  # type: ignore
                current_value=dependency_config.from_requirements_file,
                message="Cannot set from_requirements_file and from_pyproject_toml at the same time",
            )
        if any([dependency_config.pypi, dependency_config.conda]) and any(
            [
                dependency_config.from_requirements_file,
                dependency_config.from_pyproject_toml,
            ]
        ):
            raise ConfigValidationFailedException(
                field_name="pypi" if dependency_config.pypi else "conda",
                field_info=dependency_config._get_field(  # type: ignore
                    "pypi" if dependency_config.pypi else "conda"
                ),
                current_value=dependency_config.pypi or dependency_config.conda,
                message="Cannot set pypi or conda when from_requirements_file or from_pyproject_toml is set",
            )
        return True


class PackageConfig(metaclass=ConfigMeta):
    """Package configuration."""

    src_path = ConfigField(
        cli_meta=CLIOption(
            name="package_src_path",
            cli_option_str="--package-src-path",
            help="The path to the source code to deploy with the App.",
        ),
        field_type=str,
        example="./",
    )
    suffixes = ConfigField(
        cli_meta=CLIOption(
            name="package_suffixes",
            cli_option_str="--package-suffixes",
            help="A list of suffixes to add to the source code to deploy with the App.",
        ),
        field_type=list,
        example=[".py", ".ipynb"],
    )


def everything_is_string(*args):
    return all(isinstance(arg, str) for arg in args)


class BasicAppValidations:
    @staticmethod
    def name(name):
        if name is None:
            return True
        regex = r"^[a-z0-9-]+$"  # Only allow lowercase letters, numbers, and hyphens
        validator = BasicValidations(CoreConfig, "name")
        return validator.length_validation(15, name) and validator.regex_validation(
            regex, name
        )

    @staticmethod
    def port(port):
        if port is None:
            return True
        return BasicValidations(CoreConfig, "port").range_validation(1, 65535, port)

    @staticmethod
    def tags(tags):
        if tags is None:
            return True
        if not all(
            isinstance(tag, dict)
            and len(tag) == 1
            and all(
                [everything_is_string(*tag.keys()), everything_is_string(*tag.values())]
            )
            for tag in tags
        ):
            raise ConfigValidationFailedException(
                field_name="tags",
                field_info=CoreConfig._get_field(CoreConfig, "tags"),  # type: ignore
                current_value=tags,
                message="Tags must be a list of dictionaries with one key and the value must be a string. Currently they are set to %s "
                % (str(tags)),
            )
        return True

    @staticmethod
    def secrets(secrets):
        if secrets is None:  # If nothing is set we dont care.
            return True

        if not isinstance(secrets, list):
            raise ConfigValidationFailedException(
                field_name="secrets",
                field_info=CoreConfig._get_field(CoreConfig, "secrets"),  # type: ignore
                current_value=secrets,
                message="Secrets must be a list of strings. Currently they are set to %s "
                % (str(secrets)),
            )
        from ..validations import secrets_validator

        try:
            secrets_validator(secrets)
        except Exception as e:
            raise ConfigValidationFailedException(
                field_name="secrets",
                field_info=CoreConfig._get_field(CoreConfig, "secrets"),  # type: ignore
                current_value=secrets,
                message=f"Secrets validation failed, {e}",
            )
        return True

    @staticmethod
    def persistence(persistence):
        if persistence is None:
            return True
        return BasicValidations(CoreConfig, "persistence").enum_validation(
            ["none", "postgres"], persistence
        )


class CoreConfig(metaclass=ConfigMeta):
    """Unified App Configuration - The single source of truth for application configuration.

    CoreConfig is the central configuration class that defines all application settings using the
    ConfigMeta metaclass and ConfigField descriptors. It provides a declarative, type-safe way
    to manage configuration from multiple sources (CLI, config files, environment) with automatic
    validation, merging, and CLI generation.

    Core Features:
    - **Declarative Configuration**: All fields are defined using ConfigField descriptors
    - **Multi-Source Configuration**: Supports CLI options, config files (JSON/YAML), and programmatic setting
    - **Automatic CLI Generation**: CLI options are automatically generated from field metadata
    - **Type Safety**: Built-in type checking and validation for all fields
    - **Hierarchical Structure**: Supports nested configuration objects (resources, auth, dependencies)
    - **Intelligent Merging**: Configurable merging behavior for different field types
    - **Validation Framework**: Comprehensive validation with custom validation functions

    Configuration Lifecycle:
    1. **Definition**: Fields are defined declaratively using ConfigField descriptors
    2. **Instantiation**: Objects are created with all fields initialized to None or nested objects
    3. **Population**: Values are populated from CLI options, config files, or direct assignment
    4. **Merging**: Multiple config sources are merged according to field behavior settings
    5. **Validation**: Field validation functions and required field checks are performed
    6. **Default Application**: Default values are applied to any remaining None fields
    7. **Commit**: Final validation and preparation for use


    Usage Examples:
        Create from CLI options:
        ```python
        config = CoreConfig.from_cli({
            'name': 'myapp',
            'port': 8080,
            'commands': ['python app.py']
        })
        ```

        Create from config file:
        ```python
        config = CoreConfig.from_file('config.yaml')
        ```

        Create from dictionary:
        ```python
        config = CoreConfig.from_dict({
            'name': 'myapp',
            'port': 8080,
            'resources': {
                'cpu': '500m',
                'memory': '1Gi'
            }
        })
        ```

        Merge configurations:
        ```python
        file_config = CoreConfig.from_file('config.yaml')
        cli_config = CoreConfig.from_cli(cli_options)
        final_config = CoreConfig.merge_configs(file_config, cli_config)
        final_config.commit()  # Validate and apply defaults
        ```
    """

    # TODO: We can add Force Upgrade / No Deps flags here too if we need to.
    # Since those can be exposed on the CLI side and the APP state will anyways
    # be expored before being worked upon.

    SCHEMA_DOC = """Schema for defining Outerbounds Apps configuration. This schema is what we will end up using on the CLI/programmatic interface.
How to read this schema:
1. If the a property has `mutation_behavior` set to `union` then it will allow overrides of values at runtime from the CLI.
2. If the property has `mutation_behavior`set to `not_allowed` then either the CLI or the config file value will be used (which ever is not None). If the user supplies something in both then an error will be raised.
3. If a property has `experimental` set to true then a lot its validations may-be skipped and parsing handled somewhere else.
"""

    # Required fields
    name = ConfigField(
        cli_meta=CLIOption(
            name="name",
            cli_option_str="--name",
        ),
        validation_fn=BasicAppValidations.name,
        field_type=str,
        required=True,
        help="The name of the app to deploy.",
        example="myapp",
    )
    port = ConfigField(
        cli_meta=CLIOption(
            name="port",
            cli_option_str="--port",
        ),
        validation_fn=BasicAppValidations.port,
        field_type=int,
        required=True,
        help="Port where the app is hosted. When deployed this will be port on which we will deploy the app.",
        example=8000,
    )

    # Optional basic fields
    description = ConfigField(
        cli_meta=CLIOption(
            name="description",
            cli_option_str="--description",
            help="The description of the app to deploy.",
        ),
        field_type=str,
        example="This is a description of my app.",
    )
    app_type = ConfigField(
        cli_meta=CLIOption(
            name="app_type",
            cli_option_str="--app-type",
            help="The User defined type of app to deploy. Its only used for bookkeeping purposes.",
        ),
        field_type=str,
        example="MyCustomAgent",
    )
    image = ConfigField(
        cli_meta=CLIOption(
            name="image",
            cli_option_str="--image",
            help="The Docker image to deploy with the App.",
        ),
        field_type=str,
        example="python:3.10-slim",
    )

    # List fields
    tags = ConfigField(
        cli_meta=CLIOption(
            name="tags",
            cli_option_str="--tag",
            multiple=True,
            click_type=PureStringKVPairType,
        ),
        field_type=list,
        validation_fn=BasicAppValidations.tags,
        help="The tags of the app to deploy.",
        example=[{"foo": "bar"}, {"x": "y"}],
    )
    secrets = ConfigField(
        cli_meta=CLIOption(
            name="secrets", cli_option_str="--secret", multiple=True, click_type=str
        ),
        field_type=list,
        help="Outerbounds integrations to attach to the app. You can use the value you set in the `@secrets` decorator in your code.",
        example=["hf-token"],
        validation_fn=BasicAppValidations.secrets,
    )
    compute_pools = ConfigField(
        cli_meta=CLIOption(
            name="compute_pools",
            cli_option_str="--compute-pools",
            help="A list of compute pools to deploy the app to.",
            multiple=True,
            click_type=str,
        ),
        field_type=list,
        example=["default", "large"],
    )
    environment = ConfigField(
        cli_meta=CLIOption(
            name="environment",
            cli_option_str="--env",
            multiple=True,
            click_type=JsonFriendlyKeyValuePairType,  # TODO: Fix me.
        ),
        field_type=dict,
        help="Environment variables to deploy with the App.",
        example={
            "DEBUG": True,
            "DATABASE_CONFIG": {"host": "localhost", "port": 5432},
            "ALLOWED_ORIGINS": ["http://localhost:3000", "https://myapp.com"],
        },
    )
    commands = ConfigField(
        cli_meta=None,  # We dont expose commands as an options. We rather expose it like `--` with click.
        field_type=list,
        required=True,  # Either from CLI or from config file.
        help="A list of commands to run the app with.",  # TODO: Fix me: make me configurable via the -- stuff in click.
        example=["python app.py", "python app.py --foo bar"],
        behavior=FieldBehavior.NOT_ALLOWED,
    )

    # Complex nested fields
    resources = ConfigField(
        cli_meta=None,  # No top-level CLI option, only nested fields have CLI options
        field_type=ResourceConfig,
        # TODO : see if we can add a validation func for resources.
        help="Resource configuration for the app.",
    )
    auth = ConfigField(
        cli_meta=None,  # No top-level CLI option, only nested fields have CLI options
        field_type=AuthConfig,
        help="Auth related configurations.",
        validation_fn=AuthConfig.validate,
    )
    replicas = ConfigField(
        cli_meta=None,  # No top-level CLI option, only nested fields have CLI options
        validation_fn=ReplicaConfig.validate,
        field_type=ReplicaConfig,
        default=ReplicaConfig.defaults,
        help="The number of replicas to deploy the app with.",
    )
    dependencies = ConfigField(
        cli_meta=None,  # No top-level CLI option, only nested fields have CLI options
        validation_fn=DependencyConfig.validate,
        field_type=DependencyConfig,
        help="The dependencies to attach to the app. ",
    )
    package = ConfigField(
        cli_meta=None,  # No top-level CLI option, only nested fields have CLI options
        field_type=PackageConfig,
        help="Configurations associated with packaging the app.",
    )

    no_deps = ConfigField(
        cli_meta=CLIOption(
            name="no_deps",
            cli_option_str="--no-deps",
            help="Do not any dependencies. Directly used the image provided",
            is_flag=True,
        ),
        field_type=bool,
        default=False,
        help="Do not bake any dependencies. Directly used the image provided",
    )

    force_upgrade = ConfigField(
        cli_meta=CLIOption(
            name="force_upgrade",
            cli_option_str="--force-upgrade",
            help="Force upgrade the app even if it is currently being upgraded.",
            is_flag=True,
        ),
        field_type=bool,
        default=False,
        help="Force upgrade the app even if it is currently being upgraded.",
    )

    # ------- Experimental -------------
    # These options get treated in the `..experimental` module.
    # If we move any option as a first class citizen then we need to move
    # its capsule parsing from the `..experimental` module to the `..capsule.CapsuleInput` module.

    persistence = ConfigField(
        cli_meta=CLIOption(
            name="persistence",
            cli_option_str="--persistence",
            help="The persistence mode to deploy the app with.",
            choices=["none", "postgres"],
        ),
        validation_fn=BasicAppValidations.persistence,
        field_type=str,
        default="none",
        example="postgres",
        is_experimental=True,
    )

    project = ConfigField(
        cli_meta=CLIOption(
            name="project",
            cli_option_str="--project",
            help="The project name to deploy the app to.",
        ),
        field_type=str,
        is_experimental=True,
        example="my-project",
    )
    branch = ConfigField(
        cli_meta=CLIOption(
            name="branch",
            cli_option_str="--branch",
            help="The branch name to deploy the app to.",
        ),
        field_type=str,
        is_experimental=True,
        example="main",
    )
    models = ConfigField(
        cli_meta=None,
        field_type=list,
        is_experimental=True,
        example=[{"asset_id": "model-123", "asset_instance_id": "instance-456"}],
    )
    data = ConfigField(
        cli_meta=None,
        field_type=list,
        is_experimental=True,
        example=[{"asset_id": "data-789", "asset_instance_id": "instance-101"}],
    )
    # ------- /Experimental -------------

    def to_dict(self):
        return config_meta_to_dict(self)

    @staticmethod
    def merge_configs(
        base_config: "CoreConfig", override_config: "CoreConfig"
    ) -> "CoreConfig":
        """
        Merge two configurations with override taking precedence.

        Handles FieldBehavior for proper merging:
        - UNION: Merge values (for lists, dicts)
        - NOT_ALLOWED: Base config value takes precedence (override is ignored)

        Args:
            base_config: Base configuration (lower precedence)
            override_config: Override configuration (higher precedence)

        Returns:
            Merged CoreConfig instance
        """
        merged_config = CoreConfig()

        # Process each field according to its behavior
        for field_name, field_info in CoreConfig._fields.items():  # type: ignore
            base_value = getattr(base_config, field_name, None)
            override_value = getattr(override_config, field_name, None)

            # Get the behavior for this field
            behavior = getattr(field_info, "behavior", FieldBehavior.UNION)

            merged_value = merge_field_values(
                base_value, override_value, field_info, behavior
            )

            setattr(merged_config, field_name, merged_value)

        return merged_config

    def set_defaults(self):
        apply_defaults(self)

    def validate(self):
        validate_config_meta(self)

    @commit_owner_names_across_tree
    def commit(self):
        self.validate()
        validate_required_fields(self)
        self.set_defaults()

    @classmethod
    def from_dict(cls, config_data: Dict[str, Any]) -> "CoreConfig":
        config = cls()
        # Define functions for dict source
        def get_dict_key(field_name, field_info):
            return field_name

        def get_dict_value(source_data, key):
            return source_data.get(key)

        populate_config_recursive(
            config, cls, config_data, get_dict_key, get_dict_value
        )
        return config

    @classmethod
    def from_cli(cls, cli_options: Dict[str, Any]) -> "CoreConfig":
        config = cls()
        # Define functions for CLI source
        def get_cli_key(field_name, field_info):
            # Need to have a special Exception for commands since the Commands
            # are passed down via unprocessed args after `--` in click
            if field_name == cls.commands.name:
                return field_name
            # Return the CLI parameter name if CLI metadata exists
            if field_info.cli_meta and not field_info.cli_meta.hidden:
                return field_info.cli_meta.name
            return None

        def get_cli_value(source_data, key):
            value = source_data.get(key)
            # Only return non-None values since None means not set in CLI
            if value is None:
                return None
            if key == cls.environment.name:
                _env_dict = {}
                for v in value:
                    _env_dict.update(v)
                return _env_dict
            if type(value) == tuple or type(value) == list:
                obj = list(x for x in source_data[key])
                if len(obj) == 0:
                    return None  # Dont return Empty Lists so that we can set Nones
                return obj
            return value

        # Use common recursive population function with nested value checking
        populate_config_recursive(
            config,
            cls,
            cli_options,
            get_cli_key,
            get_cli_value,
        )
        return config



import functools
from typing import Optional

from metaflow import current
from metaflow.decorators import StepDecorator

from .s3_proxy_manager import S3ProxyManager
from .exceptions import S3ProxyException
from .constants import S3_PROXY_WRITE_MODES


class S3ProxyDecorator(StepDecorator):
    """
    S3 Proxy decorator for routing S3 requests through a local proxy service.

    Parameters
    ----------
    integration_name : str, optional
        Name of the S3 proxy integration. If not specified, will use the only
        available S3 proxy integration in the namespace (fails if multiple exist).
    write_mode : str, optional
        The desired behavior during write operations to target (origin) S3 bucket.
        allowed options are:
            "origin-and-cache" -> write to both the target S3 bucket and local object
                storage
            "origin" -> only write to the target S3 bucket
            "cache" -> only write to the object storage service used for caching
    debug : bool, optional
        Enable debug logging for proxy operations.
    """

    name = "s3_proxy"
    defaults = {
        "integration_name": None,
        "write_mode": None,
        "debug": False,
    }

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        write_mode = self.attributes["write_mode"]
        if write_mode and write_mode not in S3_PROXY_WRITE_MODES:
            raise S3ProxyException(
                f"unexpected write_mode specified: {write_mode}. Allowed values are: {','.join(S3_PROXY_WRITE_MODES)}."
            )

        self.manager = S3ProxyManager(
            integration_name=self.attributes["integration_name"],
            write_mode=self.attributes["write_mode"],
            debug=self.attributes["debug"],
        )

        current._update_env({"s3_proxy": self.manager})

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        """Setup S3 proxy before step execution"""
        self.manager.setup_proxy()

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_retries
    ):
        """Cleanup S3 proxy after step execution"""
        if self.manager:
            self.manager.cleanup()


class NebiusS3ProxyDecorator(S3ProxyDecorator):
    """
    Nebius-specific S3 Proxy decorator for routing S3 requests through a local proxy service.
    It exists to make it easier for users to know that this decorator should only be used with
    a Neo Cloud like Nebius.
    """

    name = "nebius_s3_proxy"
    defaults = {
        "integration_name": None,
        "write_mode": None,
        "debug": False,
    }


class CoreWeaveS3ProxyDecorator(S3ProxyDecorator):
    """
    CoreWeave-specific S3 Proxy decorator for routing S3 requests through a local proxy service.
    It exists to make it easier for users to know that this decorator should only be used with
    a Neo Cloud like CoreWeave.
    """

    name = "coreweave_s3_proxy"
    defaults = {
        "integration_name": None,
        "write_mode": None,
        "debug": False,
    }



import threading
import time
import sys
from typing import Dict, Optional, Any, Callable
from functools import partial
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import FAST_BAKERY_URL

from .fast_bakery import FastBakery, FastBakeryApiResponse, FastBakeryException
from .docker_environment import cache_request

BAKERY_METAFILE = ".imagebakery-cache"


class BakerException(MetaflowException):
    headline = "Ran into an error while baking image"

    def __init__(self, msg):
        super(BakerException, self).__init__(msg)


def bake_image(
    cache_file_path: str,
    ref: Optional[str] = None,
    python: Optional[str] = None,
    pypi_packages: Optional[Dict[str, str]] = None,
    conda_packages: Optional[Dict[str, str]] = None,
    base_image: Optional[str] = None,
    logger: Optional[Callable[[str], Any]] = None,
) -> FastBakeryApiResponse:
    """
    Bakes a Docker image with the specified dependencies.

    Args:
        cache_file_path: Path to the cache file
        ref: Reference identifier for this bake (for logging purposes)
        python: Python version to use
        pypi_packages: Dictionary of PyPI packages and versions
        conda_packages: Dictionary of Conda packages and versions
        base_image: Base Docker image to use
        logger: Optional logger function to output progress

    Returns:
        FastBakeryApiResponse: The response from the bakery service

    Raises:
        BakerException: If the baking process fails
    """
    # Default logger if none provided
    if logger is None:
        logger = partial(print, file=sys.stderr)

    # Thread lock for logging
    logger_lock = threading.Lock()
    images_baked = 0

    @cache_request(cache_file_path)
    def _cached_bake(
        ref=None,
        python=None,
        pypi_packages=None,
        conda_packages=None,
        base_image=None,
    ):
        try:
            bakery = FastBakery(url=FAST_BAKERY_URL)
            bakery._reset_payload()
            bakery.python_version(python)
            bakery.pypi_packages(pypi_packages)
            bakery.conda_packages(conda_packages)
            bakery.base_image(base_image)
            # bakery.ignore_cache()

            with logger_lock:
                logger(f"🍳 Baking [{ref}] ...")
                logger(f"     🐍 Python: {python}")

                if pypi_packages:
                    logger(f"     📦 PyPI packages:")
                    for package, version in pypi_packages.items():
                        logger(f"        🔧 {package}: {version}")

                if conda_packages:
                    logger(f"     📦 Conda packages:")
                    for package, version in conda_packages.items():
                        logger(f"        🔧 {package}: {version}")

                logger(f"     🏗️  Base image: {base_image}")

            start_time = time.time()
            res = bakery.bake()
            # TODO: Get actual bake time from bakery
            bake_time = time.time() - start_time

            with logger_lock:
                logger(f"🏁 Baked [{ref}] in {bake_time:.2f} seconds!")
            nonlocal images_baked
            images_baked += 1
            return res
        except FastBakeryException as ex:
            raise BakerException(f"Bake [{ref}] failed: {str(ex)}")

    # Call the cached bake function with the provided parameters
    return _cached_bake(
        ref=ref,
        python=python,
        pypi_packages=pypi_packages,
        conda_packages=conda_packages,
        base_image=base_image,
    )



"""
The purpose of this file is a little bit of cleverness to allow us to use the CLI in this package across
metaflow and outerbounds projects.

The issue is that since outerbounds and metaflow both vendor click, we can't use object from one import path
and expect them to work with objects created from the other import path.

Meaning `outerbounds._vendor.click.Group` and `metaflow._vendor.click.Group` are different classes.
So we need to ensure that based on when the import is taking place, we import the correct class.

Overall, this ONLY affects constructs in click we are using to construct related to the  cli decorators but
it doesn't affect any capabilities in click for logging.
"""
import os

# Import Hacks
if os.environ.get("APPS_CLI_LOADING_IN_OUTERBOUNDS", None):
    from outerbounds._vendor import click as outerbounds_click

    click = outerbounds_click  # type: ignore
else:
    from metaflow._vendor import click as metaflow_click

    click = metaflow_click  # type: ignore



from .unified_config import CoreConfig
from .cli_generator import auto_cli_options
from .config_utils import (
    PureStringKVPairType,
    JsonFriendlyKeyValuePairType,
    CommaSeparatedListType,
    MergingNotAllowedFieldsException,
    ConfigValidationFailedException,
    RequiredFieldMissingException,
)
from . import schema_export
from .typed_configs import TypedCoreConfig, TypedDict



from setuptools import setup, find_namespace_packages
from pathlib import Path


version = "1.4.10"
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    include_package_data=True,  # allow Extracting the JSON/YAML files specified in the manifest.in File.
    name="ob_metaflow_extensions",
    version=version,
    description="Outerbounds Platform Extensions for Metaflow",
    author="Outerbounds, Inc.",
    license="Commercial",
    packages=find_namespace_packages(include=["metaflow_extensions.*"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "boto3",
        "kubernetes",
        "ob-metaflow == 2.18.1.1",
    ],
)



"""
This library is an abstraction layer for connecting to snowflake using Outerbounds
OIDC tokens. It expects that a security integration that authenticates tokens minted
by Outerbounds has already been configured in the target snowflake account.
"""
from metaflow.metaflow_config import SERVICE_URL
from metaflow.metaflow_config_funcs import init_config
from typing import Dict
from os import environ
import sys
import json
import requests
import random
import time


class OuterboundsSnowflakeConnectorException(Exception):
    pass


class OuterboundsSnowflakeIntegrationSpecApiResponse:
    def __init__(self, response):
        self.response = response

    @property
    def account(self):
        return self.response["account"]

    @property
    def user(self):
        return self.response["user"]

    @property
    def default_role(self):
        return self.response["default_role"]

    @property
    def warehouse(self):
        return self.response["warehouse"]

    @property
    def database(self):
        return self.response["database"]


def get_snowflake_token(user: str = "", role: str = "", integration: str = "") -> str:
    """
    Uses the Outerbounds source token to request for a snowflake compatible OIDC
    token. This token can then be used to connect to snowflake.
    user: str
        The user the token will be minted for
    role: str
        The role to which the token will be scoped to
    integration: str
        The name of the snowflake integration to use. If not set, an existing integration will be used provided that only one exists per perimeter. If integration is not set and more than one exists, then we raise an exception.
    """
    provisioner = SnowflakeIntegrationProvisioner(integration)
    if not user or not role or not integration:
        integration_spec = provisioner.get_snowflake_integration_spec()
        if not user:
            user = integration_spec.user

        if not role:
            role = integration_spec.default_role

        if not integration:
            integration = provisioner.get_integration_name()

    snowflake_token_url = provisioner.get_snowflake_token_url()
    perimeter = provisioner.get_perimeter()
    payload = {
        "perimeterName": perimeter,
        "snowflakeUser": user,
        "snowflakeRole": role,
        "integrationName": integration,
    }
    json_payload = json.dumps(payload)
    headers = provisioner.get_service_auth_header()
    response = _api_server_get(
        snowflake_token_url, data=json_payload, headers=headers, conn_error_retries=5
    )
    response.raise_for_status()
    return response.json()["token"]


def connect(user: str = "", role: str = "", integration: str = "", **kwargs):
    """
    Connect to snowflake using the token minted by Outerbounds
    user: str
        The user name used to authenticate with snowflake
    role: str
        The role to request when connect with snowflake
    integration: str
        The name of the snowflake integration to use. If not set, an existing integration will be used provided that only one exists in the current perimeter. If integration is not set and more than one exists in the current perimeter, then we raise an exception.
    kwargs: dict
        Additional arguments to pass to the python snowflake connector
    """
    # ensure password is not set
    if "password" in kwargs:
        raise OuterboundsSnowflakeConnectorException(
            "Password should not be set when using Outerbounds snowflake connector."
        )

    provisioner = SnowflakeIntegrationProvisioner(integration)
    get_defaults = any(
        key not in kwargs for key in ["account", "warehouse", "database"]
    )
    if not user or not role or not integration or get_defaults:
        integration_spec = provisioner.get_snowflake_integration_spec()
        if not user:
            user = integration_spec.user

        if not role:
            role = integration_spec.default_role

        if not integration:
            integration = provisioner.get_integration_name()

        if "account" not in kwargs:
            kwargs["account"] = integration_spec.account

        if "warehouse" not in kwargs:
            kwargs["warehouse"] = integration_spec.warehouse

        # if the user is attempting to use a warehouse different from what is specified in the
        # integration we will not set the database
        if (
            "database" not in kwargs
            and kwargs["warehouse"] == integration_spec.warehouse
        ):
            kwargs["database"] = integration_spec.database

    # get snowflake token
    token = get_snowflake_token(user=user, role=role, integration=integration)
    kwargs["token"] = token
    kwargs["authenticator"] = "oauth"
    kwargs["role"] = role
    kwargs["user"] = user

    # connect to snowflake
    try:
        from snowflake.connector import connect

        cn = connect(**kwargs)
        return cn
    except ImportError as ie:
        raise OuterboundsSnowflakeConnectorException(
            f"Error importing snowflake connector: {ie}.\nPlease make sure the 'snowflake-connector-python' package has been installed by running 'pip install -U \"outerbounds[snowflake]\"' or using the Metaflow decorators @pypi or @conda."
        )
    except Exception as e:
        raise OuterboundsSnowflakeConnectorException(
            f"Error connecting to snowflake: {e}"
        )


def _api_server_get(*args, conn_error_retries=2, **kwargs):
    """
    There are two categories of errors that we need to handle when dealing with any API server.
    1. HTTP errors. These are are errors that are returned from the API server.
        - How to handle retries for this case will be application specific.
    2. Errors when the API server may not be reachable (DNS resolution / network issues)
        - In this scenario, we know that something external to the API server is going wrong causing the issue.
        - Failing pre-maturely in the case might not be the best course of action since critical user jobs might crash on intermittent issues.
        - So in this case, we can just planely retry the request.

    This function handles the second case. It's a simple wrapper to handle the retry logic for connection errors.
    If this function is provided a `conn_error_retries` of 5, then the last retry will have waited 32 seconds.
    Generally this is a safe enough number of retries after which we can assume that something is really broken. Until then,
    there can be intermittent issues that would resolve themselves if we retry gracefully.
    """
    _num_retries = 0
    noise = random.uniform(-0.5, 0.5)
    while _num_retries < conn_error_retries:
        try:
            return requests.get(*args, **kwargs)
        except requests.exceptions.ConnectionError:
            if _num_retries <= conn_error_retries - 1:
                # Exponential backoff with 2^(_num_retries+1) seconds
                time.sleep((2 ** (_num_retries + 1)) + noise)
                _num_retries += 1
            else:
                print(
                    "[@snowflake] Failed to connect to the API server. ",
                    file=sys.stderr,
                )
                raise


class Snowflake:
    def __init__(
        self, user: str = "", role: str = "", integration: str = "", **kwargs
    ) -> None:
        self.cn = connect(user, role, integration, **kwargs)

    def __enter__(self):
        return self.cn

    def __exit__(self, exception_type, exception_value, traceback):
        self.cn.close()

    def close(self):
        self.cn.close()


class SnowflakeIntegrationProvisioner:
    def __init__(self, integration_name: str) -> None:
        self.conf = init_config()
        self.integration_name = integration_name

    def get_snowflake_integration_spec(
        self,
    ) -> OuterboundsSnowflakeIntegrationSpecApiResponse:
        integrations_url = self._get_integration_url()
        perimeter = self.get_perimeter()
        headers = {"Content-Type": "application/json", "Connection": "keep-alive"}
        request_payload = {
            "perimeter_name": perimeter,
        }
        # if integration is not set, list all integrations
        if not self.integration_name:
            list_snowflake_integrations_url = f"{integrations_url}/snowflake"
            response = self._make_request(
                list_snowflake_integrations_url, headers, request_payload
            )
            snowflake_integrations = response.get("integrations", [])
            if not snowflake_integrations:
                raise OuterboundsSnowflakeConnectorException(
                    "No snowflake integrations found. Please make sure you have created a Snowflake integration on the Outerbounds UI first."
                )

            if len(snowflake_integrations) > 1:
                raise OuterboundsSnowflakeConnectorException(
                    f"Multiple snowflake integrations found. Please specify a specific integration name you would like to use."
                )

            self.integration_name = snowflake_integrations[0]["integration_name"]
            return OuterboundsSnowflakeIntegrationSpecApiResponse(
                snowflake_integrations[0]["integration_spec"]
            )

        get_snowflake_integration_url = (
            f"{integrations_url}/snowflake/{self.integration_name}"
        )
        response = self._make_request(
            get_snowflake_integration_url, headers, request_payload
        )
        self.integration_name = response["integration_name"]
        return OuterboundsSnowflakeIntegrationSpecApiResponse(
            response["integration_spec"]
        )

    def get_integration_name(self) -> str:
        return self.integration_name

    def get_perimeter(self) -> str:
        if "OBP_PERIMETER" in self.conf:
            perimeter = self.conf["OBP_PERIMETER"]
        else:
            # if the perimeter is not in metaflow config, try to get it from the environment
            perimeter = environ.get("OBP_PERIMETER", "")
        if not perimeter:
            raise OuterboundsSnowflakeConnectorException(
                "No perimeter set. Please make sure to run `outerbounds configure <...>` command which can be found on the Ourebounds UI or reach out to your Outerbounds support team."
            )

        return perimeter

    def get_snowflake_token_url(self) -> str:
        if "OBP_AUTH_SERVER" in self.conf:
            auth_host = self.conf["OBP_AUTH_SERVER"]
        else:
            from urllib.parse import urlparse

            auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

        return "https://" + auth_host + "/generate/snowflake"

    def get_service_auth_header(self) -> str:
        if "METAFLOW_SERVICE_AUTH_KEY" in self.conf:
            return {"x-api-key": self.conf["METAFLOW_SERVICE_AUTH_KEY"]}
        else:
            return json.loads(environ.get("METAFLOW_SERVICE_HEADERS"))

    def _get_integration_url(self) -> str:
        from metaflow_extensions.outerbounds.remote_config import init_config
        from os import environ

        if "OBP_INTEGRATIONS_URL" in self.conf:
            integrations_url = self.conf["OBP_INTEGRATIONS_URL"]
        else:
            # if the integrations url is not in metaflow config, try to get it from the environment
            integrations_url = environ.get("OBP_INTEGRATIONS_URL", "")

        if not integrations_url:
            raise OuterboundsSnowflakeConnectorException(
                "No integrations url set. Please notify your Outerbounds support team about this issue."
            )

        return integrations_url

    def _make_request(self, url, headers: Dict, payload: Dict) -> Dict:
        try:
            from metaflow.metaflow_config import SERVICE_HEADERS

            request_headers = {**headers, **(SERVICE_HEADERS or {})}
        except ImportError:
            headers = headers

        retryable_status_codes = [409]
        json_payload = json.dumps(payload)
        for attempt in range(2):  # 0 = initial attempt, 1-2 = retries
            response = _api_server_get(
                url, data=json_payload, headers=request_headers, conn_error_retries=5
            )
            if response.status_code not in retryable_status_codes:
                break

            if attempt < 2:  # Don't sleep after the last attempt
                sleep_time = 0.5 * (attempt + 1)
                time.sleep(sleep_time)

        response = _api_server_get(
            url, data=json_payload, headers=request_headers, conn_error_retries=5
        )
        self._handle_error_response(response)
        return response.json()

    @staticmethod
    def _handle_error_response(response: requests.Response):
        if response.status_code >= 500:
            raise OuterboundsSnowflakeConnectorException(
                f"Server error: {response.text}. Please reach out to your Outerbounds support team."
            )

        body = response.json()
        status_code = body.get("error", {}).get("statusCode", response.status_code)
        if status_code == 404:
            raise OuterboundsSnowflakeConnectorException(f"Secret not found: {body}")

        if status_code >= 400:
            try:
                raise OuterboundsSnowflakeConnectorException(
                    f"status_code={status_code}\t*{body['error']['details']['kind']}*\n{body['error']['details']['message']}"
                )
            except KeyError:
                raise OuterboundsSnowflakeConnectorException(
                    f"status_code={status_code} Unexpected error: {body}"
                )



import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..app_config import AppConfig

DEFAULT_BRANCH = "test"


# Account for project / branch and the capsule input.
def capsule_input_overrides(app_config: "AppConfig", capsule_input: dict):
    project = app_config.get_state("project", None)
    # Update the project/branch related configurations.
    if project is not None:
        branch = app_config.get_state("branch", DEFAULT_BRANCH)
        capsule_input["tags"].extend(
            [dict(key="project", value=project), dict(key="branch", value=branch)]
        )

    persistence = app_config.get_state("persistence", None)
    if persistence is not None and persistence != "none":
        capsule_input["persistence"] = persistence

    model_asset_conf = app_config.get_state("models", None)
    data_asset_conf = app_config.get_state("data", None)
    code_info = _code_info(app_config)
    # todo:fix me
    _objects_key = "associatedObjects"
    if model_asset_conf or data_asset_conf or code_info:
        capsule_input[_objects_key] = {}

        if model_asset_conf:
            capsule_input[_objects_key]["models"] = [
                {"assetId": x["asset_id"], "assetInstanceId": x["asset_instance_id"]}
                for x in model_asset_conf
            ]
        if data_asset_conf:
            capsule_input[_objects_key]["data"] = [
                {"assetId": x["asset_id"], "assetInstanceId": x["asset_instance_id"]}
                for x in data_asset_conf
            ]
        if code_info:
            capsule_input[_objects_key]["code"] = code_info

    return capsule_input


def _code_info(app_config: "AppConfig"):
    from metaflow.metaflow_git import get_repository_info

    try:
        from metaflow.metaflow_git import _call_git  # type: ignore
    except ImportError:
        # Fallback if _call_git is not available
        def _call_git(args, path=None):
            return "", 1, True

    repo_info = get_repository_info(app_config.get_state("packaging_directory", None))
    if len(repo_info) == 0:
        return None

    git_log_info, returncode, failed = _call_git(
        ["log", "-1", "--pretty=%B"],
        path=app_config.get_state("packaging_directory", None),
    )
    repo_url = repo_info["repo_url"]
    if isinstance(repo_url, str):
        _url = repo_url if not repo_url.endswith(".git") else repo_url.rstrip(".git")
    else:
        _url = str(repo_url)
    _code_info = {
        "commitId": repo_info["commit_sha"],
        "commitLink": os.path.join(_url, "commit", str(repo_info["commit_sha"])),
    }
    if not failed and returncode == 0 and isinstance(git_log_info, str):
        _code_info["commitMessage"] = git_log_info.strip()

    return _code_info



from typing import Dict, Optional
import requests
import time


class FastBakeryException(Exception):
    pass


class SolverStats:
    def __init__(self, stats) -> None:
        self.stats = stats

    @property
    def duration_ms(self):
        return self.stats["durationMs"]

    @property
    def packages_in_solved_environment(self):
        return self.stats["packagesInSolvedEnvironment"]


class BakingStats:
    def __init__(self, stats) -> None:
        self.stats = stats

    @property
    def solver_stats(self) -> Optional[SolverStats]:
        if "solverStats" not in self.stats:
            return None
        return SolverStats(self.stats["solverStats"])


class FastBakeryApiResponse:
    def __init__(self, response) -> None:
        self.response = response

    @property
    def python_path(self) -> Optional[str]:
        if not self.success:
            return None

        return self.response["success"]["pythonPath"]

    @property
    def container_image(self) -> Optional[str]:
        if not self.success:
            return None

        return self.response["success"]["containerImage"]

    @property
    def success(self) -> bool:
        return "success" in self.response

    @property
    def baking_stats(self) -> Optional[BakingStats]:
        if not self.success:
            return None

        if "bakingStats" not in self.response["success"]:
            return None

        if self.response["success"]["bakingStats"] is None:
            return None

        return BakingStats(self.response["success"]["bakingStats"])

    @property
    def failure(self) -> bool:
        return "failure" in self.response


class FastBakery:
    def __init__(self, url: str):
        if not url:
            raise FastBakeryException("Specifying a url is required.")
        self.url = url
        self.headers = {"Content-Type": "application/json", "Connection": "keep-alive"}
        self._reset_payload()

    def _reset_payload(self):
        self._payload = {}
        from metaflow_extensions.outerbounds.remote_config import init_config
        from os import environ

        conf = init_config()
        if "OBP_PERIMETER" in conf:
            perimeter = conf["OBP_PERIMETER"]
        else:
            # if the perimeter is not in metaflow config, try to get it from the environment
            perimeter = environ.get("OBP_PERIMETER", "")

        self._payload["perimeterName"] = perimeter

    def python_version(self, version: str):
        self._payload["pythonVersion"] = version
        return self

    def pypi_packages(self, packages: Dict[str, str]):
        self._payload.setdefault("pipRequirements", []).extend(
            self._format_packages(packages)
        )
        return self

    def conda_packages(self, packages: Dict[str, str]):
        self._payload.setdefault("condaMatchspecs", []).extend(
            self._format_packages(packages)
        )
        return self

    def base_image(self, image: str):
        self._payload["baseImage"] = {"imageReference": image}
        return self

    def image_kind(self, kind: str):
        self._payload["imageKind"] = kind
        return self

    def ignore_cache(self):
        self._payload["cacheBehavior"] = {
            "responseMaxAgeSeconds": 0,
            "layerMaxAgeSeconds": 0,
            "baseImageMaxAgeSeconds": 0,
            "overwriteExistingLayers": True,  # Used primarily to rewrite possibly corrupted layers.
        }
        return self

    @staticmethod
    def _format_packages(packages: Dict[str, str]) -> list:
        if not packages:
            return []

        def format_package(pkg: str, ver: str) -> str:
            return (
                f"{pkg}{ver}"
                if any(ver.startswith(c) for c in [">", "<", "~", "@", "="])
                else (f"{pkg}=={ver}" if ver != "" else f"{pkg}")
            )

        return [format_package(pkg, ver) for pkg, ver in packages.items()]

    def bake(self) -> FastBakeryApiResponse:
        if "imageKind" not in self._payload:
            self._payload["imageKind"] = "oci-zstd"  # Set default if not specified

        res = self._make_request(self._payload)
        self._reset_payload()
        return res

    def _make_request(self, payload: Dict) -> FastBakeryApiResponse:
        try:
            from metaflow.metaflow_config import SERVICE_HEADERS

            headers = {**self.headers, **(SERVICE_HEADERS or {})}
        except ImportError:
            headers = self.headers

        retryable_status_codes = [409]

        for attempt in range(2):  # 0 = initial attempt, 1-2 = retries
            response = requests.post(self.url, json=payload, headers=headers)

            if response.status_code not in retryable_status_codes:
                break

            if attempt < 2:  # Don't sleep after the last attempt
                sleep_time = 0.5 * (attempt + 1)
                time.sleep(sleep_time)

        response = requests.post(self.url, json=payload, headers=headers)
        self._handle_error_response(response)
        return FastBakeryApiResponse(response.json())

    @staticmethod
    def _handle_error_response(response: requests.Response):
        if response.status_code >= 500:
            raise FastBakeryException(f"Server error: {response.text}")

        body = response.json()
        status_code = body.get("error", {}).get("statusCode", response.status_code)
        if status_code >= 400:
            try:
                raise FastBakeryException(
                    f"*{body['error']['details']['kind']}*\n{body['error']['details']['message']}"
                )
            except KeyError:
                raise FastBakeryException(f"Unexpected error: {body}")



from metaflow.user_decorators.mutable_flow import MutableFlow
from metaflow.user_decorators.mutable_step import MutableStep
from metaflow.user_decorators.user_flow_decorator import FlowMutator
from .assume_role import OBP_ASSUME_ROLE_ARN_ENV_VAR


class assume_role(FlowMutator):
    """
    Flow-level decorator for assuming AWS IAM roles.

    When applied to a flow, all steps in the flow will automatically use the specified IAM role-arn
    as their source principal.

    Usage:
    ------
    @assume_role(role_arn="arn:aws:iam::123456789012:role/my-iam-role")
    class MyFlow(FlowSpec):
        @step
        def start(self):
            import boto3
            client = boto3.client("dynamodb")  # Automatically uses the role in the flow decorator
            self.next(self.end)

        @step
        def end(self):
            from metaflow import get_aws_client
            client = get_aws_client("dynamodb")  # Automatically uses the role in the flow decorator
    """

    def init(self, *args, **kwargs):
        self.role_arn = kwargs.get("role_arn", None)

        if self.role_arn is None:
            raise ValueError(
                "`role_arn` keyword argument is required for the assume_role decorator"
            )

        if not self.role_arn.startswith("arn:aws:iam::"):
            raise ValueError(
                "`role_arn` must be a valid AWS IAM role ARN starting with 'arn:aws:iam::'"
            )

    def pre_mutate(self, mutable_flow: MutableFlow) -> None:
        """
        This method is called by Metaflow to apply the decorator to the flow.
        It sets up environment variables that will be used by the AWS client
        to automatically assume the specified role.
        """
        # Import environment decorator at runtime to avoid circular imports
        from metaflow import environment

        def _swap_environment_variables(step: MutableStep, role_arn: str) -> None:
            _step_has_env_set = True
            _env_kwargs = {OBP_ASSUME_ROLE_ARN_ENV_VAR: role_arn}
            for d in step.decorator_specs:
                name, _, _, deco_kwargs = d
                if name == "environment":
                    _env_kwargs.update(deco_kwargs["vars"])
                    _step_has_env_set = True

            if _step_has_env_set:
                # remove the environment decorator
                step.remove_decorator("environment")

            # add the environment decorator
            step.add_decorator(
                environment,
                deco_kwargs=dict(vars=_env_kwargs),
            )

        # Set the role ARN as an environment variable that will be picked up
        # by the get_aws_client function
        def _setup_role_assumption(step: MutableStep) -> None:
            _swap_environment_variables(step, self.role_arn)

        # Apply the role assumption setup to all steps in the flow
        for _, step in mutable_flow.steps:
            _setup_role_assumption(step)



from metaflow.exception import MetaflowException


class NvctExecutionException(MetaflowException):
    headline = "Nvct task couldn't be executed"


class NvctTaskFailedException(MetaflowException):
    headline = "Nvct task failed"


class NvctKilledException(MetaflowException):
    headline = "Nvct job killed"


class RequestedGPUTypeUnavailableException(MetaflowException):
    headline = "[@nvct RequestedGPUTypeUnavailableException] GPU type unavailable."

    def __init__(self, requested_gpu_type, available_gpus):
        msg = (
            f"The requested GPU type @nvct(..., gpu_type='{requested_gpu_type}') is not available. "
            f"Please choose from the following supported GPU types when using @nvct: {available_gpus}"
        )
        super(RequestedGPUTypeUnavailableException, self).__init__(msg)


class UnsupportedNvctConfigurationException(MetaflowException):
    headline = (
        "[@nvct UnsupportedNvctConfigurationException] Unsupported GPU configuration"
    )

    def __init__(self, n_gpu, gpu_type, available_configurations, step):
        msg = f"The requested configuration of @nvct(gpu={n_gpu}, gpu_type='{gpu_type}') for @step {step} is not available."
        if len(available_configurations) == 0:
            msg += (
                "\n\nNo configurations are available in your Outerbounds deployment."
                " Please contact Outerbounds support if you wish to use @nvct."
            )
        else:
            msg += f"\n\nAvailable configurations for your deployment with {gpu_type} include: \n\t- {self._display(gpu_type, available_configurations)}"
            msg += "\n\nPlease contact Outerbounds support if you wish to use a configuration not listed above."
        super(UnsupportedNvctConfigurationException, self).__init__(msg)

    def _display(self, gpu_type, configs):
        _available_decos = []
        for cfg in configs:
            n_gpu = cfg["n_gpus"]
            _available_decos.append(f"@nvct(gpu={n_gpu}, gpu_type='{gpu_type}')")
        return "\n\t- ".join(_available_decos)


class UnsupportedNvctDatastoreException(MetaflowException):
    headline = "[@nvct UnsupportedNvctDatastoreException] Unsupported datastore"

    def __init__(self, ds_type):
        msg = (
            "The *@nvct* decorator requires --datastore=s3 or --datastore=azure or --datastore=gs at the moment."
            f"Current datastore type: {ds_type}."
        )
        super(UnsupportedNvctDatastoreException, self).__init__(msg)


class NvctTimeoutTooShortException(MetaflowException):
    headline = "[@nvct NvctTimeoutTooShortException] Timeout too short"

    def __init__(self, step):
        msg = (
            "The timeout for step *{step}* should be at least 60 seconds for "
            "execution with @nvct".format(step=step)
        )
        super(NvctTimeoutTooShortException, self).__init__(msg)



from . import app_cli
from . import config
from .deployer import AppDeployer, apps
from .config.typed_configs import (
    ReplicaConfigDict,
    ResourceConfigDict,
    AuthConfigDict,
    DependencyConfigDict,
    PackageConfigDict,
)



from metaflow.exception import MetaflowException


class UnspecifiedRemoteStorageRootException(MetaflowException):
    headline = "Storage root not specified."

    def __init__(self, message):
        super(UnspecifiedRemoteStorageRootException, self).__init__(message)


class EmptyOllamaManifestCacheException(MetaflowException):
    headline = "Model not found."

    def __init__(self, message):
        super(EmptyOllamaManifestCacheException, self).__init__(message)


class EmptyOllamaBlobCacheException(MetaflowException):
    headline = "Blob not found."

    def __init__(self, message):
        super(EmptyOllamaBlobCacheException, self).__init__(message)



from metaflow.user_decorators.mutable_flow import MutableFlow
from .external_chckpt import _ExternalCheckpointFlowDeco
import os

NEBIUS_ENDPOINT_URL = "https://storage.eu-north1.nebius.cloud:443"


class nebius_checkpoints(_ExternalCheckpointFlowDeco):

    """

    This decorator is used for setting the nebius's S3 compatible object store as the artifact store for
    checkpoints/models created by the flow.

    Parameters
    ----------
    secrets: list
        A list of secrets to be added to the step. These secrets should contain any secrets that are required globally and the secret
        for the nebius object store. The secret should contain the following keys:
        - NEBIUS_ACCESS_KEY
        - NEBIUS_SECRET_KEY

    bucket_path: str
        The path to the bucket to store the checkpoints/models.

    endpoint_url: str
        The endpoint url for the nebius object store. Defaults to `https://storage.eu-north1.nebius.cloud:443`

    Usage
    -----
    ```python
    from metaflow import checkpoint, step, FlowSpec, nebius_checkpoints

    @nebius_checkpoints(secrets=[], bucket_path=None)
    class MyFlow(FlowSpec):
        @checkpoint
        @step
        def start(self):
            # Saves the checkpoint in the nebius object store
            current.checkpoint.save("./foo.txt")

        @step
        def end(self):
            pass
    ```
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.nebius_endpoint_url = kwargs.get("endpoint_url", NEBIUS_ENDPOINT_URL)

    def pre_mutate(self, mutable_flow: MutableFlow) -> None:
        from metaflow import (
            with_artifact_store,
        )

        def _nebius_config():
            return {
                "root": self.bucket_path,
                "client_params": {
                    "aws_access_key_id": os.environ.get("NEBIUS_ACCESS_KEY"),
                    "aws_secret_access_key": os.environ.get("NEBIUS_SECRET_KEY"),
                    "endpoint_url": self.nebius_endpoint_url,
                },
            }

        mutable_flow.add_decorator(
            with_artifact_store, deco_kwargs=dict(type="s3", config=_nebius_config)
        )
        self._swap_secrets(mutable_flow)



from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow import current
from .core import AppDeployer, apps
from .core.perimeters import PerimeterExtractor
import os
import hashlib


class AppDeployDecorator(StepDecorator):

    """
    MF Add To Current
    -----------------
    apps -> metaflow_extensions.outerbounds.plugins.apps.core.apps

        @@ Returns
        ----------
        apps
            The object carrying the Deployer class to deploy apps.
    """

    name = "app_deploy"
    defaults = {}

    package_url = None
    package_sha = None

    MAX_ENTROPY = 6
    MAX_NAME_LENGTH = 15 - MAX_ENTROPY - 1  # -1 for the hyphen

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

    def _resolve_package_url_and_sha(self):
        return os.environ.get("METAFLOW_CODE_URL", self.package_url), os.environ.get(
            "METAFLOW_CODE_SHA", self.package_sha
        )

    def _extract_project_info(self):
        project = current.get("project_name")
        branch = current.get("branch_name")
        is_production = current.get("is_production")
        return project, branch, is_production

    def _resolve_default_image(self, flow):
        # TODO : Resolve the default image over here.
        pass

    def _resolve_default_name_prefix(self, flow, step_name):
        # TODO: Only tweek MAX_NAME_LENGTH as backend support allows longer names.
        base_prefix = (flow.name + "-" + step_name).lower()
        if len(base_prefix) > self.MAX_NAME_LENGTH:
            base_prefix = "mf-app"
        return base_prefix

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        perimeter, api_server = PerimeterExtractor.during_metaflow_execution()
        package_url, package_sha = self._resolve_package_url_and_sha()
        if package_url is None or package_sha is None:
            raise MetaflowException(
                "METAFLOW_CODE_URL or METAFLOW_CODE_SHA is not set. "
                "Please set METAFLOW_CODE_URL and METAFLOW_CODE_SHA in your environment."
            )
        image = os.environ.get("FASTBAKERY_IMAGE", None)

        # TODO [Apps] - This is temporary. Backend will support longer names in the future.
        default_name = self._resolve_default_name_prefix(flow, step_name)
        project, branch, is_production = self._extract_project_info()
        project_info = {}
        if project is not None:
            project_info["metaflow/project"] = project
            project_info["metaflow/branch"] = branch
            project_info["metaflow/is_production"] = is_production

        default_tags = {
            "metaflow/flow_name": flow.name,
            "metaflow/step_name": step_name,
            "metaflow/run_id": run_id,
            "metaflow/task_id": task_id,
            "metaflow/retry_count": retry_count,
            "metaflow/pathspec": current.pathspec,
            **project_info,
        }

        AppDeployer._set_state(
            perimeter,
            api_server,
            code_package_url=package_url,
            code_package_key=package_sha,
            name_prefix=default_name,
            image=image,
            max_entropy=self.MAX_ENTROPY,
            default_tags=[{k: str(v)} for k, v in default_tags.items()],
        )
        current._update_env(
            {
                "apps": apps(),
            }
        )

    def task_post_step(
        self, step_name, flow, graph, retry_count, max_user_code_retries
    ):
        pass

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        # To execute the Kubernetes job, the job container needs to have
        # access to the code package. We store the package in the datastore
        # which the pod is able to download as part of it's entrypoint.
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]
            os.environ["METAFLOW_CODE_URL"] = cls.package_url
            os.environ["METAFLOW_CODE_SHA"] = cls.package_sha



from datetime import datetime
from metaflow.decorators import StepDecorator
from ..card_utilities.injector import CardDecoratorInjector


class DynamicCardAppendDecorator(StepDecorator):
    """
    A simple decorator that demonstrates using CardDecoratorInjector
    to inject a card and render simple markdown content.
    """

    name = "test_append_card"

    defaults = {
        "title": "Simple Card",
        "message": "Hello from DynamicCardAppendDecorator!",
        "show_timestamp": True,
        "refresh_interval": 5,
    }

    CARD_ID = "simple_card"

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        """Initialize the decorator and inject the card."""
        self.deco_injector = CardDecoratorInjector()
        self.deco_injector.attach_card_decorator(
            flow,
            step_name,
            self.CARD_ID,
            "blank",
            refresh_interval=self.attributes["refresh_interval"],
        )

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        """Decorate the step function to add card content."""
        from metaflow import current
        from metaflow.cards import Markdown

        # Create the card content
        title = self.attributes["title"]
        message = self.attributes["message"]
        show_timestamp = self.attributes["show_timestamp"]

        # Add title to the card
        current.card[self.CARD_ID].append(Markdown(f"# {title}"))

        # Add message to the card
        current.card[self.CARD_ID].append(Markdown(f"**Message:** {message}"))

        # Add timestamp if requested
        if show_timestamp:
            timestamp = datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S %z")
            current.card[self.CARD_ID].append(Markdown(f"**Created at:** {timestamp}"))

        # Add step information
        current.card[self.CARD_ID].append(Markdown(f"**Step:** `{current.pathspec}`"))

        # Add a simple divider
        current.card[self.CARD_ID].append(Markdown("---"))

        # Add some dynamic content that shows this is working
        current.card[self.CARD_ID].append(
            Markdown("**Status:** Card successfully injected! 🎉")
        )

        def wrapped_step_func():
            """Execute the original step function."""
            try:
                # Before execution
                current.card[self.CARD_ID].append(
                    Markdown("**Execution:** Step started...")
                )
                current.card[self.CARD_ID].refresh()

                # Execute the original step
                step_func()

                # After execution
                current.card[self.CARD_ID].append(
                    Markdown("**Execution:** Step completed successfully! ✅")
                )
                current.card[self.CARD_ID].refresh()

            except Exception as e:
                # Handle errors
                current.card[self.CARD_ID].append(
                    Markdown(f"**Error:** Step failed with error: `{str(e)}` ❌")
                )
                current.card[self.CARD_ID].refresh()
                raise

        return wrapped_step_func



import os
import sys

from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator
from metaflow.plugins.parallel_decorator import ParallelDecorator
from metaflow.metadata_provider.util import sync_local_metadata_to_datastore
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.sidecar import Sidecar
from metaflow.plugins.timeout_decorator import get_run_time_limit_for_task
from metaflow.metadata_provider import MetaDatum

from .utils import get_ngc_api_key
from .exceptions import (
    UnsupportedNvctDatastoreException,
    NvctTimeoutTooShortException,
    RequestedGPUTypeUnavailableException,
    UnsupportedNvctConfigurationException,
)


DEFAULT_GPU_TYPE = "H100"

SUPPORTABLE_GPU_TYPES = {
    "L40": [
        {
            "n_gpus": 1,
            "instance_type": "gl40_1.br20_2xlarge",
            "backend": "GFN",
        },
    ],
    "L40S": [
        {
            "n_gpus": 1,
            "instance_type": "gl40s_1.br25_2xlarge",
            "backend": "GFN",
        },
    ],
    "L40G": [
        {
            "n_gpus": 1,
            "instance_type": "gl40g_1.br25_2xlarge",
            "backend": "GFN",
        },
    ],
    "H100": [
        {
            "n_gpus": 1,
            "instance_type": "OCI.GPU.H100_1x",
            "backend": "nvcf-dgxc-k8s-oci-nrt-prd8",
        },
        {
            "n_gpus": 2,
            "instance_type": "OCI.GPU.H100_2x",
            "backend": "nvcf-dgxc-k8s-oci-nrt-prd8",
        },
        {
            "n_gpus": 4,
            "instance_type": "OCI.GPU.H100_4x",
            "backend": "nvcf-dgxc-k8s-oci-nrt-prd8",
        },
        {
            "n_gpus": 8,
            "instance_type": "OCI.GPU.H100_8x",
            "backend": "nvcf-dgxc-k8s-oci-nrt-prd8",
        },
    ],
    "NEBIUS_H100": [
        {
            "n_gpus": 1,
            "instance_type": "ON-PREM.GPU.H100_1x",
            "backend": "default-project-eu-north1",
        },
        {
            "n_gpus": 2,
            "instance_type": "ON-PREM.GPU.H100_2x",
            "backend": "default-project-eu-north1",
        },
        {
            "n_gpus": 4,
            "instance_type": "ON-PREM.GPU.H100_4x",
            "backend": "default-project-eu-north1",
        },
        {
            "n_gpus": 8,
            "instance_type": "ON-PREM.GPU.H100_8x",
            "backend": "default-project-eu-north1",
        },
    ],
}


class NvctDecorator(StepDecorator):

    """
    Specifies that this step should execute on DGX cloud.

    Parameters
    ----------
    gpu : int
        Number of GPUs to use.
    gpu_type : str
        Type of Nvidia GPU to use.
    """

    name = "nvct"
    defaults = {
        "gpu": 1,
        "gpu_type": None,
        "ngc_api_key": None,
        "instance_type": None,
        "backend": None,
    }

    package_url = None
    package_sha = None

    # Refer https://github.com/Netflix/metaflow/blob/master/docs/lifecycle.png
    # to understand where these functions are invoked in the lifecycle of a
    # Metaflow flow.
    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # Executing NVCT functions requires a non-local datastore.
        if flow_datastore.TYPE not in ("s3", "azure", "gs"):
            raise UnsupportedNvctDatastoreException(flow_datastore.TYPE)

        # Set internal state.
        self.logger = logger
        self.environment = environment
        self.step = step
        self.flow_datastore = flow_datastore

        if any([deco.name == "kubernetes" for deco in decos]):
            raise MetaflowException(
                "Step *{step}* is marked for execution both on Kubernetes and "
                "Nvct. Please use one or the other.".format(step=step)
            )
        if any([isinstance(deco, ParallelDecorator) for deco in decos]):
            raise MetaflowException(
                "Step *{step}* contains a @parallel decorator "
                "with the @nvct decorator. @parallel decorators are not currently supported with @nvct.".format(
                    step=step
                )
            )

        # Set run time limit for NVCT.
        self.run_time_limit = get_run_time_limit_for_task(decos)
        if self.run_time_limit < 60:
            raise NvctTimeoutTooShortException(step)

        self.attributes["ngc_api_key"] = get_ngc_api_key()

        requested_gpu_type = self.attributes["gpu_type"]
        requested_n_gpus = self.attributes["gpu"]

        if requested_gpu_type is None:
            requested_gpu_type = DEFAULT_GPU_TYPE
        if requested_gpu_type not in SUPPORTABLE_GPU_TYPES:
            raise RequestedGPUTypeUnavailableException(
                requested_gpu_type, list(SUPPORTABLE_GPU_TYPES.keys())
            )

        valid_config = None
        available_configurations = SUPPORTABLE_GPU_TYPES[requested_gpu_type]
        for each_config in available_configurations:
            if each_config["n_gpus"] == requested_n_gpus:
                valid_config = each_config
                break

        if valid_config is None:
            raise UnsupportedNvctConfigurationException(
                requested_n_gpus,
                requested_gpu_type,
                available_configurations,
                step,
            )

        self.attributes["instance_type"] = valid_config["instance_type"]
        self.attributes["gpu_type"] = requested_gpu_type
        if self.attributes["gpu_type"] == "NEBIUS_H100":
            self.attributes["gpu_type"] = "H100"
        self.attributes["backend"] = valid_config["backend"]

    def runtime_init(self, flow, graph, package, run_id):
        # Set some more internal state.
        self.flow = flow
        self.graph = graph
        self.package = package
        self.run_id = run_id

    def runtime_task_created(
        self, task_datastore, task_id, split_index, input_paths, is_cloned, ubf_context
    ):
        if not is_cloned:
            self._save_package_once(self.flow_datastore, self.package)

    def runtime_step_cli(
        self, cli_args, retry_count, max_user_code_retries, ubf_context
    ):
        if retry_count <= max_user_code_retries:
            # after all attempts to run the user code have failed, we don't need
            # to execute on NVCF anymore. We can execute possible fallback
            # code locally.
            cli_args.commands = ["nvct", "step"]
            cli_args.command_args.append(self.package_sha)
            cli_args.command_args.append(self.package_url)
            cli_options = {
                "gpu_type": self.attributes["gpu_type"],
                "instance_type": self.attributes["instance_type"],
                "backend": self.attributes["backend"],
                "ngc_api_key": self.attributes["ngc_api_key"],
            }
            cli_args.command_options.update(cli_options)
            cli_args.entrypoint[0] = sys.executable

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_retries,
        ubf_context,
        inputs,
    ):
        self.metadata = metadata
        self.task_datastore = task_datastore

        # task_pre_step may run locally if fallback is activated for @catch
        # decorator.

        if "NVCT_CONTEXT" in os.environ:
            meta = {}

            meta["nvct-task-id"] = os.environ.get("NVCT_TASK_ID")
            meta["nvct-task-name"] = os.environ.get("NVCT_TASK_NAME")
            meta["nvct-ncaid"] = os.environ.get("NVCT_NCA_ID")
            meta["nvct-progress-file-path"] = os.environ.get("NVCT_PROGRESS_FILE_PATH")
            meta["nvct-results-dir"] = os.environ.get("NVCT_RESULTS_DIR")

            entries = [
                MetaDatum(
                    field=k,
                    value=v,
                    type=k,
                    tags=["attempt_id:{0}".format(retry_count)],
                )
                for k, v in meta.items()
                if v is not None
            ]
            # Register book-keeping metadata for debugging.
            metadata.register_metadata(run_id, step_name, task_id, entries)

            self._save_logs_sidecar = Sidecar("save_logs_periodically")
            self._save_logs_sidecar.start()

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_retries
    ):
        # task_finished may run locally if fallback is activated for @catch
        # decorator.
        if "NVCT_CONTEXT" in os.environ:
            # If `local` metadata is configured, we would need to copy task
            # execution metadata from the NVCT container to user's
            # local file system after the user code has finished execution.
            # This happens via datastore as a communication bridge.
            if hasattr(self, "metadata") and self.metadata.TYPE == "local":
                sync_local_metadata_to_datastore(
                    DATASTORE_LOCAL_DIR, self.task_datastore
                )

        try:
            self._save_logs_sidecar.terminate()
        except:
            # Best effort kill
            pass

    @classmethod
    def _save_package_once(cls, flow_datastore, package):
        if cls.package_url is None:
            cls.package_url, cls.package_sha = flow_datastore.save_data(
                [package.blob], len_hint=1
            )[0]



from metaflow.exception import MetaflowException


class S3ProxyException(MetaflowException):
    headline = "S3 Proxy Error"


class S3ProxyConfigException(S3ProxyException):
    headline = "S3 Proxy Configuration Error"


class S3ProxyApiException(S3ProxyException):
    headline = "S3 Proxy API Error"
